#############################################################################
# Copyright (c) 2024 László Várady
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 2 as published
# by the Free Software Foundation, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
# As an additional exemption you are allowed to compile & link against the
# OpenSSL libraries as published by the OpenSSL project. See the file
# COPYING for details.
#
#############################################################################

from .source import HTTPSource

__all__ = [
    "HTTPSource"
]



#############################################################################
# Copyright (c) 2015-2016 Balabit
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 2 as published
# by the Free Software Foundation, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
# As an additional exemption you are allowed to compile & link against the
# OpenSSL libraries as published by the OpenSSL project. See the file
# COPYING for details.
#
#############################################################################

from .completerlang import CompleterLang
from .commandlinelexer import CommandLineLexer
from .getoptlexer import GetoptLexer


class DebugLang(CompleterLang):
    # we only need to list commands we want to automatically complete arguments.
    # Commands that have no arguments don't need to be listed as they wouldn't really
    # have rules anyway.
    _known_commands = (
        "print",
        "display",
    )
    _aliases = {
        'p': 'print',
    }
    tokens = [
        "COMMAND_PRINT", "COMMAND_DISPLAY",
        "COMMAND",
        "ARG",
    ]

    def p_statement(self, p):
        '''statement : print_command
                    | display_command
                    | generic_command'''

    def p_args(self, p):
        '''args : ARG args
               | '''

    def p_print_command(self, p):
        '''print_command : COMMAND_PRINT template'''

    def p_display_command(self, p):
        '''display_command : COMMAND_DISPLAY template'''

    def p_generic_command(self, p):
        '''generic_command : COMMAND args'''

    def p_template(self, p):
        '''template : ARG'''

    def _construct_lexer(self):
        return GetoptLexer(CommandLineLexer(),
                           known_commands=self._known_commands,
                           aliases=self._aliases)



#############################################################################
# Copyright (c) 2015-2016 Balabit
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 2 as published
# by the Free Software Foundation, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
# As an additional exemption you are allowed to compile & link against the
# OpenSSL libraries as published by the OpenSSL project. See the file
# COPYING for details.
#
#############################################################################
# pylint: disable=import-outside-toplevel, unused-import

from __future__ import absolute_import


def is_readline_available():
    try:
        import readline
    except ImportError:
        return False
    return True


def is_editline_available():
    try:
        import editline
    except ImportError:
        return False
    return True


def setup_read_or_editline():
    if is_readline_available():
        from syslogng.debuggercli.readline import setup_readline
        setup_readline()
    elif is_editline_available():
        from syslogng.debuggercli.editline import setup_editline
        setup_editline()


def fetch_command():
    setup_read_or_editline()
    try:
        input_function = raw_input
    except NameError:
        input_function = input
    return input_function("(syslog-ng) ")



#############################################################################
# Copyright (c) 2015-2016 Balabit
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 2 as published
# by the Free Software Foundation, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
# As an additional exemption you are allowed to compile & link against the
# OpenSSL libraries as published by the OpenSSL project. See the file
# COPYING for details.
#
#############################################################################

from __future__ import print_function, absolute_import
from abc import ABCMeta, abstractmethod


class Lexer():
    __metaclass__ = ABCMeta

    @abstractmethod
    def token(self):
        raise NotImplementedError

    @abstractmethod
    def input(self, text):
        raise NotImplementedError

    @abstractmethod
    def get_position(self):
        raise NotImplementedError



#############################################################################
# Copyright (c) 2015-2016 Balabit
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 2 as published
# by the Free Software Foundation, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
# As an additional exemption you are allowed to compile & link against the
# OpenSSL libraries as published by the OpenSSL project. See the file
# COPYING for details.
#
#############################################################################

from __future__ import absolute_import, print_function
from .completer import Completer
from .syslognginternals import get_nv_registry


class MacroCompleter(Completer):
    _date_wildcards = {
        'R_': 'received time',
        'S_': 'stamp in message',
        'C_': 'current time'
    }

    def __init__(self, macros=None):
        super().__init__()
        self._macros = macros
        self._completions = []

    def complete(self, entire_text, word_to_be_completed):
        if not self._looks_like_a_macro_prefix(entire_text):
            return []
        self._collect_completions(word_to_be_completed)
        return sorted([completion
                       for completion in self._completions
                       if len(word_to_be_completed) == 0 or completion.startswith(word_to_be_completed)])

    def _looks_like_a_macro_prefix(self, entire_input):
        if entire_input == '':
            return True
        if entire_input == '$':
            return True
        if entire_input[0] == '$' and self._is_valid_macro(entire_input[1:]):
            return True
        if entire_input == '${':
            return True
        if entire_input[:2] == '${':
            return True
        return False

    def _is_valid_macro(self, macro):
        return all(self._is_valid_macro_char(c) for c in macro)

    @staticmethod
    def _is_valid_macro_char(macro_char):
        if 'A' <= macro_char <= 'Z':
            return True
        if 'a' <= macro_char <= 'z':
            return True
        if '0' <= macro_char <= '9':
            return True
        if macro_char == '_':
            return True
        return False

    def _collect_completions(self, word):
        self._reset_completions()
        if word == '':
            self._add_completion('$')
            self._add_completion('${')
        elif word == '$':
            self._extend_completions(self._collect_unqualified_macros())
            self._extend_completions(self._collect_qualified_macros_with_brace())
            self._extend_completions(self._collect_small_numbered_matches())
            self._extend_completions(self._collect_date_wildcards())
            self._add_completion('${')
        elif self._is_word_a_numbered_match_prefix(word):
            self._extend_completions(self._collect_all_numbered_matches())
        elif self._is_word_a_nonbraced_prefix(word):
            if self._is_word_a_date_prefix(word):
                self._extend_completions(self._collect_date_macros())
            else:
                self._extend_completions(self._collect_unqualified_macros())
                self._extend_completions(self._collect_date_wildcards())
        elif word == '${':
            self._extend_completions(self._collect_small_numbered_matches_with_brace())
            self._extend_completions(self._collect_all_macros_with_brace())
            self._extend_completions(self._collect_date_wildcard_with_brace())
        elif self._is_word_a_numbered_match_prefix_with_brace(word):
            self._extend_completions(self._collect_all_numbered_matches_with_brace())
        elif self._is_word_a_braced_prefix(word):
            if self._is_word_a_date_prefix_with_brace(word):
                self._extend_completions(self._collect_date_macros_with_brace())
            else:
                self._extend_completions(self._collect_all_macros_with_brace())
                self._extend_completions(self._collect_date_wildcard_with_brace())

        return self._completions

    def _is_word_a_date_prefix(self, word):
        return word[:3] in ['$' + x for x in self._date_wildcards]

    @staticmethod
    def _is_word_a_numbered_match_prefix(word):
        return word[0] == '$' and word[1:2].isdigit()

    @staticmethod
    def _is_word_a_nonbraced_prefix(word):
        return word[0] == '$' and word[1] != '{'

    @staticmethod
    def _is_word_a_numbered_match_prefix_with_brace(word):
        return word[0:2] == '${' and word[2:3].isdigit()

    def _is_word_a_date_prefix_with_brace(self, word):
        return word[:4] in ['${' + x for x in self._date_wildcards]

    @staticmethod
    def _is_word_a_braced_prefix(word):
        return word[0:2] == '${'

    @staticmethod
    def _is_macro_qualified(macro):
        return '.' in macro

    @classmethod
    def _is_macro_unqualified(cls, macro):
        return (not cls._is_macro_qualified(macro) and
                not cls._is_macro_a_numbered_match(macro) and
                not cls._is_macro_a_date_macro(macro))

    @staticmethod
    def _is_macro_a_numbered_match(macro):
        return macro.isdigit()

    @classmethod
    def _is_macro_a_date_macro(cls, macro):
        return macro[0:2] in cls._date_wildcards

    @classmethod
    def _is_macro_a_small_numbered_match(cls, macro):
        return cls._is_macro_a_numbered_match(macro) and int(macro) < 10

    def _collect_macros_generic(self, predicate, template):
        for macro in self._macros or get_nv_registry():
            if predicate(macro):
                yield template.format(macro)

    def _collect_macros(self, predicate):
        return self._collect_macros_generic(predicate, '${}')

    def _collect_macros_with_brace(self, predicate):
        return self._collect_macros_generic(predicate, '${{{}}}')

    def _collect_unqualified_macros(self):
        return self._collect_macros(self._is_macro_unqualified)

    def _collect_qualified_macros_with_brace(self):
        return self._collect_macros_with_brace(self._is_macro_qualified)

    def _collect_date_macros(self):
        return self._collect_macros(self._is_macro_a_date_macro)

    def _collect_date_macros_with_brace(self):
        return self._collect_macros_with_brace(self._is_macro_a_date_macro)

    def _collect_small_numbered_matches(self):
        return self._collect_macros(self._is_macro_a_small_numbered_match)

    def _collect_small_numbered_matches_with_brace(self):
        return self._collect_macros_with_brace(self._is_macro_a_small_numbered_match)

    def _collect_all_numbered_matches(self):
        return self._collect_macros(self._is_macro_a_numbered_match)

    def _collect_all_numbered_matches_with_brace(self):
        return self._collect_macros_with_brace(self._is_macro_a_numbered_match)

    def _collect_all_macros_with_brace(self):
        return self._collect_macros_with_brace(
            lambda macro: (not self._is_macro_a_numbered_match(macro) and
                           not self._is_macro_a_date_macro(macro)))

    def _collect_date_wildcards(self):
        for (wildcard, description) in self._date_wildcards.items():
            yield f"${wildcard}* ({description})"

    def _collect_date_wildcard_with_brace(self):
        for (wildcard, description) in self._date_wildcards.items():
            yield f"${{{wildcard}*}} ({description})"

    def _reset_completions(self):
        self._completions = []

    def _extend_completions(self, iterable):
        self._completions.extend(iterable)

    def _add_completion(self, value):
        self._completions.append(value)



#############################################################################
# Copyright (c) 2022 Balazs Scheidler <bazsi77@gmail.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
# As an additional exemption you are allowed to compile & link against the
# OpenSSL libraries as published by the OpenSSL project. See the file
# COPYING for details.
#
#############################################################################
# pylint: disable=function-redefined

try:
    from _syslogng import LogDestination, LogDestinationResult

except ImportError:
    import warnings
    from enum import Enum, auto

    warnings.warn("You have imported the syslogng package outside of syslog-ng, "
                  "thus some of the functionality is not available. "
                  "Defining fake classes for those exported by the underlying syslog-ng code")

    LogDestination = object

    class LogDestinationResult(Enum):
        DROP = auto()
        ERROR = auto()
        SUCCESS = auto()
        QUEUED = auto()
        NOT_CONNECTED = auto()


class LogDestination(LogDestination):
    DROP = LogDestinationResult.DROP
    ERROR = LogDestinationResult.ERROR
    SUCCESS = LogDestinationResult.SUCCESS
    QUEUED = LogDestinationResult.QUEUED
    NOT_CONNECTED = LogDestinationResult.NOT_CONNECTED

    def open(self):
        """Open a connection to the target service"""
        return True

    def close(self):
        """Close the connection to the target service"""

    def is_opened(self):
        """Check if the connection to the target is able to receive messages"""
        return True

    def init(self, options):
        """Initialize this log element

        This method is called when the syslog-ng configuration is
        initialized.  You can prevent the configuration to be accepted by
        returning False, in which case you should log why that was the case.

        The same object can be initalized multiple times in case the
        configuration reload operation fails in syslog-ng:
          1) init() when the config is first initialized
          2) deinit() as a preparation for reload
          3) initialization of the new config fails for some reason (any one
             object can cause it to fail)
          4) init() again, as syslog-ng reverts to the old config in such cases.

        Args:
            options:
                is a dictionary that contains options passed using the
                options() parameter of the python() destination

        Returns:
            bool: True if successful, False otherwise
        """
        return True

    def deinit(self):
        """Deinitialize this log element

        This method is called when the syslog-ng configuration is being shut
        down.
        """

    def send(self, msg):
        """Send a message to the target service

        This method is invoked by the underlying C code to dispatch a
        message to the target service.

        Args:
            msg (LogMessage): message to be sent

        Returns:
            int: one value from the LogDestinationResult enum
        """
        raise NotImplementedError



#############################################################################
# Copyright (c) 2025 One Identity LLC
# Copyright (c) 2023 Attila Szakacs
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 2 as published
# by the Free Software Foundation, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
# As an additional exemption you are allowed to compile & link against the
# OpenSSL libraries as published by the OpenSSL project. See the file
# COPYING for details.
#
#############################################################################
from boto3.s3.transfer import TransferConfig
from boto3.session import Session
from boto3 import client
from botocore.credentials import DeferredRefreshableCredentials, create_assume_role_refresher
from botocore.exceptions import ClientError, EndpointConnectionError

from .s3_object_buffer import S3ObjectQueue, S3ObjectBuffer

from concurrent.futures import ThreadPoolExecutor, Future, CancelledError
from logging import getLogger
from typing import Any, Dict, Set
from threading import Lock
import re


class S3SessionHandler:

    logger = getLogger("S3.SessionHandler")
    __max_retries = 10

    def __init__(self, max_concurrent_uploads: int, upload_threads: int, object_queue: S3ObjectQueue, bucket: str, role: str | None, s3_session_config: Dict[str, Any], s3_client_config: Dict[str, Any], s3_sse_config: Dict[str, Any], s3_transfer_config: TransferConfig) -> None:
        self.__bucket = bucket
        self.__role = role
        self.__client_config = s3_client_config
        if s3_sse_config.get("ServerSideEncryption", None) is not None:
            self.__client_config.update(s3_sse_config)
        self.__session = Session(**s3_session_config)
        self.__sts = None
        self.__client = None

        # NOTE: The Session.set_credentials always creates a new Credentials object from the given keys.
        # NOTE: The DeferredRefreshableCredentials class is a child of RefreshableCredentials which is a
        # NOTE: child of the Credentials class.
        # ALSO NOTE: This solution is ugly as hell. We should find a more elegant one.
        if role:
            self.__sts = client("sts")
            self.__session._session._credentials = DeferredRefreshableCredentials(
                refresh_using=create_assume_role_refresher(
                    self.__sts,
                    {"RoleArn": self.__role, "RoleSessionName": "syslog-ng"},
                ),
                method="sts-assume-role",
            )
            self.logger.info(f"Using {self.__sts.get_caller_identity().get('Arn')} for the current session.")

        self.__transfer_config = s3_transfer_config
        self.__object_queue = object_queue
        self.__max_concurrent_uploads = max_concurrent_uploads
        self.__upload_threads = upload_threads
        self.__thread_pool = ThreadPoolExecutor(max_workers=max_concurrent_uploads)
        self.__futures: set[Future] = set()
        self.__future_lock = Lock()
        self.__connection_lock = Lock()
        self.__upload_enabled = True

    def open_connection(self):
        with self.__connection_lock:
            self.__client = self.__session.client("s3", **self.__client_config)

    def close_connection(self):
        with self.__connection_lock:
            self.__client.close()
            self.__client = None

    def disable_upload(self):
        with self.__connection_lock:
            self.__upload_enabled = False

    def enable_upload(self):
        with self.__connection_lock:
            self.__upload_enabled = True

    @property
    def connection_open(self):
        with self.__connection_lock:
            return self.__client is not None

    def ready_bucket(self):
        try:
            self.__client.head_bucket(Bucket=self.__bucket)
            return True
        except ClientError as e:
            if e.response["Error"]["Code"] == "404":
                self.logger.info(f"Bucket ({self.__bucket}) does not exist, trying to create it")
            try:
                self.__client.create_bucket(Bucket=self.__bucket)
            except (ClientError, EndpointConnectionError) as e:
                self.logger.error(f"Failed to create bucket ({self.__bucket}): {e}")
                return False
            self.logger.info(f"Bucket ({self.__bucket}) successfully created")
            return True

    def upload_object(self, s3_object: S3ObjectBuffer) -> bool:
        if not self.connection_open:
            self.logger.error(f"Connection closed. Establishing client connection.")
            self.open_connection()
        rc = False
        if s3_object is None:
            return rc
        if not self.__upload_enabled:
            return rc
        try:
            extra_args: Dict[str, str] = {}
            if s3_object.storage_class is not None:
                extra_args["StorageClass"] = s3_object.storage_class
            if s3_object.acl is not None:
                extra_args["ACL"] = s3_object.acl
            if s3_object.content_type is not None:
                extra_args["ContentType"] = s3_object.content_type
            self.__client.upload_file(Filename=s3_object.path, Bucket=self.__bucket, Key=s3_object.full_target_key, Config=self.__transfer_config, ExtraArgs=extra_args)
            rc = True
            self.logger.debug(f"Object {s3_object.full_target_key} successfully uploaded.")
            s3_object.deprecate()
        except ClientError as e:
            self.logger.error(f"Could not upload object {s3_object.path}, retries: {s3_object.retries}/{self.__max_retries}. Reason: {e}")
            s3_object.retries += 1
            if s3_object.retries < self.__max_retries:
                self.__object_queue.enqueue(s3_object, None)
                return rc
            self.logger.warning(f"Maximum numbers of retries exceeded for object {s3_object.path}")

        return rc

    def __upload_complete_cb(self, future: Future) -> None:
        try:
            future.result()
        except ClientError as e:
            self.logger.error(f"Could not upload object: {e}")
        self.__futures.remove(future)

    def upload_loop(self) -> None:
        while True:
            with self.__future_lock:
                if len(self.__futures) < self.__max_concurrent_uploads:
                    s3_object = self.__object_queue.dequeue()
                    if s3_object is None:
                        return
                    self.logger.debug(f"Available executor found for uploading object: {s3_object.full_target_key}")
                    future = self.__thread_pool.submit(self.upload_object, s3_object)
                    future.add_done_callback(self.__upload_complete_cb)
                    self.__futures.add(future)
                else:
                    return

    def trigger_upload(self) -> None:
        self.upload_loop()

    def wait_for_queue_empty(self, timeout) -> bool:
        with self.__connection_lock and self.__future_lock:
            if len(self.__futures) == 0:
                return True
            for future in set(self.__futures):
                try:
                    future.result(timeout)
                except TimeoutError:
                    self.logger.warning("Object upload timed out.")
                    if not future.cancel():
                        self.logger.error("Timed out object upload could not be cancelled.")
                except CancelledError:
                    self.logger.warning("Object upload already cancelled.")
                except Exception as e:
                    self.logger.error(f"Future could not complete. Reason: {e}")
            if len(self.__futures) == 0:
                return True
            else:
                self.logger.error("Could not finish some upload futures. Uploads will be retried on next syslog-ng startup.")
                return False

    def get_top_index_for_key(self, key: str, suffix: str) -> int | None:
        top_index = -1
        objects: Set[str] = set()
        pagination_options: Dict[str, str] = {}

        while True:
            try:
                response: Dict[str, Any] = self.__client.list_objects(
                    Bucket=self.__bucket,
                    Prefix=key,
                    **pagination_options,
                )
            except (ClientError, EndpointConnectionError) as e:
                self.logger.error(f"Failed to list objects matching prefix: {self.__bucket}/{key} => {e}")
                return None

            try:
                for obj in response.get("Contents", []):
                    objects.add(obj["Key"])
                if not response["IsTruncated"]:
                    break
                pagination_options = {
                    "Marker": response["Contents"][-1]["Key"]
                }
            except KeyError:
                self.logger.error(f"Failed to list objects: {self.__bucket}/{key} => Unexpected response: {response}")
                return None

        for s3_object in objects:
            if s3_object == key + suffix:
                key_index = 0
            else:
                key_match = re.search(rf"\d+{suffix}$", s3_object)
                if key_match:
                    key_index = int(key_match.group().removesuffix(suffix))
                else:
                    key_index = -1
            if key_index >= top_index:
                top_index = key_index

        # NOTE: python3 does not have a max integer size, thus this should not result in an overflow error
        return top_index if top_index >= 0 else None



#############################################################################
# Copyright (c) 2025 One Identity LLC
# Copyright (c) 2023 Attila Szakacs
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 2 as published
# by the Free Software Foundation, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
# As an additional exemption you are allowed to compile & link against the
# OpenSSL libraries as published by the OpenSSL project. See the file
# COPYING for details.
#
#############################################################################


try:
    from .s3_object_buffer import S3ObjectBuffer, S3ObjectQueue
    from .s3_session_handler import S3SessionHandler

    # NOTE: These are imports required to get to the part processing `deps_installed`
    from logging import getLogger
    from signal import signal, SIGINT, SIG_IGN
    from syslogng import LogDestination, LogMessage, LogTemplate, get_installation_path_for

    from importlib.metadata import version
    from os import listdir, mkdir
    from pathlib import Path
    from sys import exc_info
    from threading import Timer, Lock
    from time import time
    from typing import Any, Dict, Optional

    from boto3 import client, Session
    from boto3.s3.transfer import TransferConfig
    import botocore
    from botocore.config import Config
    from botocore.credentials import create_assume_role_refresher, DeferredRefreshableCredentials
    from botocore.exceptions import ClientError, EndpointConnectionError

    deps_missing = None
    deps_installed = True
except ImportError as import_error:
    deps_missing = import_error.name
    deps_installed = False

signal(SIGINT, SIG_IGN)


class S3Destination(LogDestination):
    S3_OBJECT_TIMEOUT_INTERVAL_SECONDS = 60
    S3_OBJECT_MIN_CHUNK_SIZE = 5 * 1024 * 1024

    logger = getLogger("S3")

    def __init_options(self, options: Dict[str, Any]) -> None:
        try:
            self.url = str(options["url"])
            self.bucket = str(options["bucket"])
            self.access_key = str(options["access_key"])
            self.secret_key = str(options["secret_key"])
            self.role = str(options["role"])
            self.object_key: LogTemplate = options["object_key"]
            self.object_key_timestamp: Optional[LogTemplate] = options["object_key_timestamp"]
            self.object_key_suffix: Optional[str] = options["object_key_suffix"]
            self.message_template: LogTemplate = options["template"]
            self.compression = bool(options["compression"])
            if self.compression:
                self.object_key_suffix += ".gz"
            self.compresslevel = int(options["compresslevel"])
            self.chunk_size = int(options["chunk_size"])
            self.max_object_size = int(options["max_object_size"])
            self.upload_threads = int(options["upload_threads"])
            self.max_pending_uploads = int(options["max_pending_uploads"])
            self.flush_grace_period = int(options["flush_grace_period"])
            self.region: Optional[str] = str(options["region"])
            self.server_side_encryption = str(options["server_side_encryption"])
            self.kms_key = str(options["kms_key"])
            self.storage_class = str(options["storage_class"]).upper().replace("-", "_")
            self.canned_acl = str(options["canned_acl"]).lower().replace("_", "-")
            self.content_type = str(options["content_type"])
            self.use_checksum = str(options["use_checksum"])
        except KeyError:
            assert False, (
                f"S3: {str(exc_info()[1])[1:-1]}() option is missing. "
                "If you are using this driver via the python() destination, please use the s3() driver directly. "
                "The option validation and propagation should be done by the s3.conf SCL."
            )

        self.__persist_name = self.generate_persist_name(options)

        if str(self.object_key_timestamp) == "":
            self.object_key_timestamp = None

        if self.compresslevel < 0 or self.compresslevel > 9:
            self.logger.warning("compresslevel() must be an integer between 0 and 9. Using 9")
            self.compresslevel = 9

        if self.chunk_size < self.S3_OBJECT_MIN_CHUNK_SIZE:
            self.logger.warning(f"chunk-size() must be at least {self.S3_OBJECT_MIN_CHUNK_SIZE}. Using minimal value")
            self.chunk_size = self.S3_OBJECT_MIN_CHUNK_SIZE

        if self.upload_threads < 1:
            self.logger.warning("upload-threads() must be a positive integer. Using 1")
            self.upload_threads = 1

        if self.max_pending_uploads < 1:
            self.logger.warning("max-pending-uploads() must be a positive integer. Using 1")
            self.max_pending_uploads = 1

        if self.flush_grace_period < 1:
            self.logger.warning("flush-grace-period() must be a positive integer. Using 1")
            self.flush_grace_period = 1

        if self.region == "":
            self.region = None

        if self.server_side_encryption != "" and self.server_side_encryption != "aws:kms":
            assert False, "server-side-encryption() supports only aws:kms"

        if self.server_side_encryption == "aws:kms" and self.kms_key == "":
            assert False, "kms-key() must be set when server-side-encryption() is aws:kms"

        if self.kms_key != "" and self.server_side_encryption == "":
            self.logger.warning("ignoring kms-key() as server-side-encryption() is disabled")
            self.kms_key = ""

        VALID_STORAGE_CLASSES = {
            "STANDARD",
            "REDUCED_REDUNDANCY",
            "STANDARD_IA",
            "ONEZONE_IA",
            "INTELLIGENT_TIERING",
            "GLACIER",
            "DEEP_ARCHIVE",
            "OUTPOSTS",
            "GLACIER_IR",
            "SNOW",
        }
        if self.storage_class not in VALID_STORAGE_CLASSES:
            self.logger.warning(
                f"Invalid storage-class(). Valid values are: f{', '.join(sorted(VALID_STORAGE_CLASSES))}. Using STANDARD"
            )
            self.storage_class = "STANDARD"

        VALID_CANNED_ACLS = {
            "private",
            "public-read",
            "public-read-write",
            "aws-exec-read",
            "authenticated-read",
            "bucket-owner-read",
            "bucket-owner-full-control",
            "log-delivery-write",
        }
        if self.canned_acl != "" and self.canned_acl not in VALID_CANNED_ACLS:
            self.logger.warning(
                f"Invalid canned-acl(). Valid values are: f{', '.join(sorted(VALID_CANNED_ACLS))} or empty. Using empty"
            )
            self.canned_acl = ""

        VALID_CHECKSUM_SETTINGS = {
            "when_supported",
            "when_required",
        }
        if self.use_checksum not in VALID_CHECKSUM_SETTINGS:
            self.logger.warning(
                f"Invalid use-checksum(). Valid values are: {', '.join(sorted(VALID_CHECKSUM_SETTINGS))} or empty. Using when_supported."
            )
            self.use_checksum = "when_supported"

    def init(self, options: Dict[str, Any]) -> bool:
        if not deps_installed:
            if deps_missing:
                self.logger.error(
                    f"Unable to start the Python based S3 destination. The required module dependency '{deps_missing}' could not be found.'"
                )
            else:
                self.logger.error(
                    "Unable to start the Python based S3 destination, some required Python dependencies (likely `boto3` and/or `botocore`) are missing"
                )
            return False

        self.__init_options(options)

        self.s3_session_config: Dict[str, Any] = {
            "aws_access_key_id": self.access_key if self.access_key != "" else None,
            "aws_secret_access_key": self.secret_key if self.secret_key != "" else None,
            "region_name": self.region,
        }
        self.s3_client_config: Dict[str, Any] = {
            "endpoint_url": self.url if self.url != "" else None,
        }
        if version('botocore') > '1.36':
            self.s3_client_config.update({
                "config": Config(
                    request_checksum_calculation=self.use_checksum,
                    response_checksum_validation=self.use_checksum,
                )
            })
        else:
            self.logger.info("The option 'use-checksum()' requires at least botocore version 1.36. Current version is {}".format(version('botocore')))
        self.s3_sse_options: Dict[str, Any] = {
            "ServerSideEncryption": self.server_side_encryption if self.server_side_encryption != "" else None,
            "SSEKMSKeyId": self.kms_key if self.server_side_encryption != "" else None,
        }
        self.transfer_config = TransferConfig(
            multipart_threshold=int(self.chunk_size * 1.5),
            max_concurrency=self.upload_threads,
            multipart_chunksize=self.chunk_size,
            use_threads=True if self.upload_threads > 1 else False,
        )
        self.object_config: Dict[str, Any] = {
            "suffix": self.object_key_suffix,
            "storage_class": self.storage_class,
            "compression": self.compression,
            "compresslevel": self.compresslevel,
            "max_object_size": self.max_object_size,
            "canned_acl": self.canned_acl,
            "content_type": self.content_type,
        }

        self.s3_object_ready_queue: S3ObjectQueue = S3ObjectQueue()
        self.s3_objects_active: Dict[str, S3ObjectBuffer] = {}
        self.__objects_lock = Lock()
        self.__indices_lock = Lock()
        self.__indices: Dict[str, int] = {}

        self.__session_handler: S3SessionHandler = S3SessionHandler(self.max_pending_uploads, self.upload_threads, self.s3_object_ready_queue, self.bucket, self.role, self.s3_session_config, self.s3_client_config, self.s3_sse_options, self.transfer_config)

        self.flush_poll_event: Optional[Timer] = None

        self.working_dir = Path(get_installation_path_for(r"${localstatedir}"), "s3")
        try:
            mkdir(self.working_dir)
        except FileExistsError:
            pass

        self.__load_uncompleted_buffers()

        return True

    def deinit(self) -> None:
        pass

    @staticmethod
    def generate_persist_name(options: Dict[str, Any]) -> str:
        return f"s3({','.join([options['url'], options['bucket'], str(options['object_key'])])})"

    def __load_uncompleted_buffers(self) -> None:
        with self.__objects_lock and self.__indices_lock:
            self.logger.info(f"Trying to load uncompleted object buffers.")
            json_list = listdir(self.working_dir)
            json_list = filter(lambda file: file.endswith(".json"), json_list)
            for s3_buffer_meta in json_list:
                s3_object_buffer = S3ObjectBuffer(self.working_dir, "", 0, "", self.__persist_name, self.object_config)
                try:
                    s3_object_buffer.load_metadata(Path(self.working_dir, s3_buffer_meta))
                except KeyError as e:
                    self.logger.info(f"File {s3_buffer_meta} does not describe an object buffer.")
                    continue
                if s3_object_buffer.creator != self.__persist_name:
                    self.logger.info(f"File buffer {s3_object_buffer.path} does not belong to this driver. Will not load.")
                    continue
                if s3_object_buffer.finished:
                    self.s3_object_ready_queue.enqueue(s3_object_buffer, None)
                    self.logger.info(f"Completed object {s3_object_buffer.path} loaded.")
                else:
                    if s3_object_buffer.object_key in self.s3_objects_active:
                        if self.s3_objects_active[s3_object_buffer.object_key].target_index < s3_object_buffer.target_index:
                            self.s3_object_ready_queue.enqueue(self.s3_objects_active[s3_object_buffer.object_key], None)
                            self.s3_objects_active[s3_object_buffer.object_key] = s3_object_buffer
                        elif self.s3_objects_active[s3_object_buffer.object_key].target_index > s3_object_buffer.target_index:
                            self.s3_object_ready_queue.enqueue(s3_object_buffer, None)
                        else:
                            self.logger.error(
                                f"Invalid indices found for object key {s3_object_buffer.object_key} "
                                f"({self.s3_objects_active[s3_object_buffer.object_key].full_target_key} "
                                f"conflicts {s3_object_buffer.full_target_key}). "
                                f"Manual intervention necessary, withholding files from upload."
                            )
                            json_list = filter(lambda file: file.startswith(s3_object_buffer.object_key), json_list)
                            self.s3_objects_active.pop(s3_object_buffer.object_key)
                            continue
                    else:
                        self.s3_objects_active[s3_object_buffer.object_key] = s3_object_buffer
                    self.logger.info(f"Incomplete object {s3_object_buffer.path} loaded")
                self.__indices[s3_object_buffer.object_key] = s3_object_buffer.target_index
        if not self.s3_object_ready_queue.queue.empty():
            self.__session_handler.trigger_upload()
        self.logger.info(f"Finished loading uncompleted object buffers.")

    def open(self) -> bool:
        self.logger.debug("Opening S3 connection.")
        if self.__session_handler.connection_open:
            self.logger.debug("S3 connection already open.")
            return True

        try:
            self.__session_handler.open_connection()

            if not self.__session_handler.ready_bucket():
                return False

            if self.flush_poll_event is None:
                self.__start_flush_poll_timer()

        except EndpointConnectionError as e:
            self.logger.error(f"Could not connect to S3 endpoint {self.url}. Reason: {e}")

        self.__session_handler.enable_upload()

        return True

    def __flush_queue(self):
        self.logger.debug("Flushing S3 queue.")
        keys = list(self.s3_objects_active.keys())
        with self.__objects_lock:
            for s3_object in keys:
                self.__finish_s3_object(self.s3_objects_active[s3_object])

        self.__session_handler.wait_for_queue_empty(10)

    def is_opened(self) -> bool:
        return self.__session_handler.connection_open

    def close(self) -> None:
        self.logger.debug(f"Closing S3 connection.")
        if not self.is_opened():
            self.logger.debug(f"S3 connection already closed.")
            return

        assert self.__session_handler.connection_open

        if self.flush_poll_event is not None:
            self.flush_poll_event.cancel()
            self.flush_poll_event = None

        self.__flush_queue()
        self.__session_handler.disable_upload()
        self.__session_handler.close_connection()

    def __start_flush_poll_timer(self) -> None:
        self.flush_poll_event = Timer(
            interval=self.S3_OBJECT_TIMEOUT_INTERVAL_SECONDS, function=self.__flush_timed_out_s3_objects
        )
        self.flush_poll_event.start()

    def __flush_timed_out_s3_objects(self) -> None:
        self.logger.debug("Flushing timed out S3 objects.")

        def should_flush_s3_object(s3_target_object: S3ObjectBuffer, current_time: float) -> bool:
            return not s3_target_object.finished and s3_target_object.last_modified + self.flush_grace_period * 60 <= current_time

        # NOTE: This used to be a monotonic value, but comparison with unix timestamps proved to be problematic
        now = int(time())

        with self.__objects_lock:
            object_list = list(self.s3_objects_active.values())
            for s3_object in object_list:
                if should_flush_s3_object(s3_object, now):
                    self.__finish_s3_object(s3_object)
                    with self.__indices_lock:
                        last_index = self.__indices.pop(s3_object.object_key)
                        self.__indices[s3_object.object_key] = last_index
        if not self.s3_object_ready_queue.queue.empty():
            self.__session_handler.trigger_upload()

        self.__start_flush_poll_timer()

    def __format_target_key(self, msg: LogMessage) -> str:
        return self.object_key.format(msg, self.template_options)

    def __format_timestamp(self, msg: LogMessage) -> str:
        if self.object_key_timestamp is not None:
            return self.object_key_timestamp.format(msg, self.template_options)
        return ""

    def __finish_s3_object(self, s3_object: S3ObjectBuffer) -> None:
        if not s3_object.finished:
            s3_object.finish()
        self.s3_objects_active.pop(s3_object.object_key)
        self.s3_object_ready_queue.enqueue(s3_object, self.__session_handler.trigger_upload)

    def __update_target_index(self, target_key):
        with self.__objects_lock and self.__indices_lock:
            greatest_online_index = self.__session_handler.get_top_index_for_key(target_key, self.object_key_suffix)
            if greatest_online_index is None:
                self.logger.error(f"Could not get latest online index for target key {target_key}. Continuing with internal indexing only.")
                greatest_online_index = -1
            if target_key in self.__indices.keys():
                self.__indices[target_key] = max(self.__indices[target_key] + 1, greatest_online_index + 1)
            else:
                self.__indices[target_key] = greatest_online_index + 1

    def __get_s3_object(self, msg: LogMessage) -> S3ObjectBuffer:
        timestamp = self.__format_timestamp(msg)
        target_key = self.__format_target_key(msg) + timestamp

        with self.__objects_lock:
            # Completely new target
            if target_key not in self.s3_objects_active:
                self.logger.debug(f"Could not fetch S3 object, creating new one.")
                self.__update_target_index(target_key)
                with self.__indices_lock:
                    s3_object = self.s3_objects_active[target_key] = S3ObjectBuffer(self.working_dir, target_key, self.__indices[target_key], timestamp, self.__persist_name, self.object_config)
                return s3_object
            else:
                return self.s3_objects_active[target_key]

    def send(self, msg: LogMessage) -> int:

        s3_object = self.__get_s3_object(msg)

        data = self.message_template.format(msg, self.template_options).encode("utf-8")

        try:
            if not s3_object.write(data):
                self.logger.error(f"Failed to write data: {data}")
        except OSError as e:
            self.logger.error(f"Failed to write data: {e}")
            return self.ERROR

        if s3_object.finished:
            # The S3 object finished itself after a successful write.
            with self.__objects_lock:
                self.__finish_s3_object(s3_object)

        self.stats_written_bytes_add(len(data))
        return self.SUCCESS
